'''Sequence class.

:copyright: 2015, Jeroen van der Heijden (Transceptor Technology)
'''
from .elements import NamedElement, c_export


class Sequence(NamedElement):

    __slots__ = ('_elements',)

    def __init__(self, *elements):
        self._elements = self._validate_elements(elements)

    def _get_node_result(self, root, tree, rule, _s, node):
        pos = node.start

        for elem in self._elements:
            is_valid, pos = root._walk(elem, pos, node.children, rule, True)
            if not is_valid:
                return is_valid, pos

        root._append_tree(tree, node, pos)

        return is_valid, pos

    def _run_export_js(self, js_identation, ident, classes):
        return self._export_js_elements(js_identation, ident, classes)

    def _run_export_py(self, py_identation, ident, classes):
        return self._export_py_elements(py_identation, ident, classes)

    def _run_export_c(self, c_identation, ident, enums):
        return self._export_c_elements(c_identation, ident, enums)


'''Exceptions.

:copyright: 2015, Jeroen van der Heijden (Transceptor Technology)
'''


class CompileError(Exception):
    pass


class KeywordError(CompileError):
    pass


class ReKeywordsChangedError(CompileError):
    pass


class NameAssignedError(CompileError):
    pass


class MissingRefError(CompileError):
    pass


'''Element and NamedElement Class.

These are the base classes used for all other elements.

:copyright: 2015, Jeroen van der Heijden (Transceptor Technology)
'''

def c_export(func):

    def wrapper(self, c_identation, ident, enums):
        gid = getattr(self, 'name', getattr(self, '_name', 'CLERI_NONE'))
        if gid != 'CLERI_NONE':
            gid = 'CLERI_GID_{}'.format(gid.upper())
            enums.add(gid)
        return func(self, c_identation, ident, enums, gid)

    return wrapper

class Element:

    __slots__ = tuple()

    @staticmethod
    def _validate_element(element):
        if isinstance(element, str):
            return Token(element)
        if isinstance(element, Element):
            return element
        raise TypeError(
            'Expecting an element or string but received type: {}'.format(
                type(element)))

    @classmethod
    def _validate_elements(cls, elements):
        return [cls._validate_element(elem) for elem in elements]


class NamedElement(Element):

    __slots__ = ('name',)

    def _export_js(self, js_identation, ident, classes):
        classes.add(self.__class__.__name__.lstrip('_'))
        if hasattr(self, 'name') and ident:
            return self.name
        return self._run_export_js(js_identation, ident or 1, classes)

    def _export_js_elements(self, js_identation, ident, classes):
        new_ident = ident + 1
        value = ',\n'.join(['{ident}{elem}'.format(
            ident=js_identation * new_ident,
            elem=elem._export_js(
                js_identation,
                new_ident, classes)) for elem in self._elements])
        return '{class_name}(\n{value}\n{ident})'.format(
            class_name=self.__class__.__name__.lstrip('_'),
            value=value,
            ident=js_identation * ident)

    def _run_export_js(self, js_identation, ident, classes):
        return 'not_implemented'

    def _export_py(self, py_identation, ident, classes):
        classes.add(self.__class__.__name__.lstrip('_'))
        if hasattr(self, 'name') and ident:
            return self.name
        return self._run_export_py(py_identation, ident or 1, classes)

    def _export_py_elements(self, py_identation, ident, classes):
        new_ident = ident + 1
        value = ',\n'.join(['{ident}{elem}'.format(
            ident=py_identation * new_ident,
            elem=elem._export_py(
                py_identation,
                new_ident, classes)) for elem in self._elements])
        return '{class_name}(\n{value}\n{ident})'.format(
            class_name=self.__class__.__name__.lstrip('_'),
            value=value,
            ident=py_identation * ident)

    def _run_export_py(self, py_identation, ident, classes):
        return 'not_implemented'

    @c_export
    def _export_c(self, c_identation, ident, enums, gid):
        if hasattr(self, 'name') and ident:
            return self.name
        return self._run_export_c(c_identation, ident or 1, enums)

    @c_export
    def _export_c_elements(self, c_identation, ident, enums, gid):
        new_ident = ident + 1
        value = ',\n'.join(['{ident}{elem}'.format(
            ident=c_identation * new_ident,
            elem=elem._export_c(
                c_identation,
                new_ident,
                enums)) for elem in self._elements])
        return 'cleri_{class_name}(\n{gid},\n{num},\n{value}\n{ident})'.format(
            class_name=self.__class__.__name__.lstrip('_').lower(),
            gid='{ident}{gid}'.format(
                ident=c_identation * (ident + 1),
                gid=gid),
            num='{ident}{num}'.format(
                ident=c_identation * (ident + 1),
                num=len(self._elements)),
            value=value,
            ident=c_identation * ident)

    def _run_export_c(self, c_identation, ident, enums):
        return 'not_implemented'

# Added this import to the bottom to prevent circular import cycle.
# Note: usually this is bad design but in this case we do want class
#       inheritance which allows us to create a new class Token which
#       is sub-classed from the 'NamedElement' class.
from .token import Token  # nopep8



'''pyleri.Ref Class.

When a forward reference is used the grammar must contain a final
reference with the same name. The Grammar class is used to validate
a valid grammer and set element.

:copyright: 2015, Jeroen van der Heijden (Transceptor Technology)
'''
from .elements import NamedElement


class Ref(NamedElement):
    '''Ref class.

    Note: _run_export_c() is not needed since C does not need a Ref but
    instead the object will be pre-declared.'''

    __slots__ = ('_element', '_get_node_result')

    @property
    def element(self):
        return getattr(self, '_element', None)

    @element.setter
    def element(self, element):
        self._element = self._validate_element(element)
        self._get_node_result = self._element._get_node_result

    def _run_export_js(self, js_identation, ident, classes):
        return 'Ref({})'.format(self._element.__class__.__name__)

    def _run_export_py(self, py_identation, ident, classes):
        return 'Ref()'



'''Regex class.

:copyright: 2015, Jeroen van der Heijden (Transceptor Technology)
'''
import re
from .elements import NamedElement, c_export


class Regex(NamedElement):

    __slots__ = ('_compiled',)

    def __init__(self, pattern, flags=0):
        if not pattern.startswith('^'):
            pattern = '^' + pattern
        self._compiled = re.compile(pattern, flags=flags)

    def __repr__(self):
        name = getattr(self, 'name', None)
        return self.__class__.__name__ if name is None else '"{}"'.format(name)

    def _get_node_result(self, root, tree, rule, s, node):
        re_match = self._compiled.match(s)
        is_valid = bool(re_match)

        if is_valid:
            root._append_tree(tree, node, node.start + len(re_match.group(0)))
        else:
            root._expecting.update(self, node.start)

        return is_valid, node.end or node.start

    def _run_export_js(self, js_identation, ident, classes):
        return 'Regex(\'{}\')'.format(
            self._compiled.pattern.replace('\\', '\\\\').replace('\'', '\\\''))

    def _run_export_py(self, py_identation, ident, classes):
        return 'Regex(\'{}\')'.format(
            self._compiled.pattern.replace('\\', '\\\\').replace('\'', '\\\''))

    @c_export
    def _run_export_c(self, c_identation, ident, enums, gid):
        return 'cleri_regex({}, "{}")'.format(
            gid,
            self._compiled.pattern
                .replace('\\', '\\\\')
                .replace('\'', '\\\'')
                .replace('\\"', '"')
                .replace('"', '\\"'))



'''Py-LeRi (pyleri) Python LR-parsing module.

This module is inspired by lrparsing (http://lrparsing.sourceforge.net/),
a Python parser written by Russell Stuart, 2014-05-29.

We did however found some strange, conflict behavior using this module and
therefore decided to build our own parser which works together with the
Javascript JS-LeRi (jsleri) module so we can exchange grammars written by this
module with JavaScript projects.

:copyright: 2016, Jeroen van der Heijden (Transceptor Technology)
:license: MIT
'''

from .choice import Choice
from .grammar import Grammar
from .keyword import Keyword
from .list import List
from .optional import Optional
from .prio import Prio
from .ref import Ref
from .regex import Regex
from .repeat import Repeat
from .sequence import Sequence
from .this import THIS
from .token import Token
from .tokens import Tokens
from .exceptions import (
    CompileError,
    KeywordError,
    ReKeywordsChangedError,
    NameAssignedError)

__author__ = 'Jeroen van der Heijden'
__maintainer__ = 'Jeroen van der Heijden'
__email__ = 'jeroen@transceptor.technology'
__version__ = '1.1.3'



'''pyleri.Grammar Class.

When creating a new grammar this class should be used as the base class.

Example:

class MyGrammar(Grammar):
    ...


:copyright: 2015, Jeroen van der Heijden (Transceptor Technology)
'''


import re
import time
from .node import Node
from .expecting import Expecting
from .endofstatement import end_of_statement
from .elements import Element, NamedElement
from .keyword import Keyword
from .ref import Ref
from .noderesult import NodeResult
from .exceptions import (
    KeywordError,
    ReKeywordsChangedError,
    NameAssignedError,
    MissingRefError)


_RE_KEYWORDS = re.compile('^\w+')


class _KeepOrder(dict):

    def __init__(self, *args):
        self._order = []
        self._refs = {}
        self._RE_KEYWORDS = _RE_KEYWORDS
        self._has_keywords = False
        super().__setitem__('_order', self._order)
        super().__setitem__('_refs', self._refs)

    def _check_keywords(self, element):
        if isinstance(element, Keyword):
            self._has_keywords = True
            m = self._RE_KEYWORDS.match(element._keyword)
            if m is None or m.group(0) != element._keyword:
                raise KeywordError(
                    'Keyword {} does not match Grammars keywords match'
                    .format(element._keyword))
        elif hasattr(element, '_elements'):
            for elem in element._elements:
                self._check_keywords(elem)

    def __setitem__(self, key, value):
        if key not in self:
            self._order.append(key)
        elif key in self._refs:
            self._refs[key].element = value
            if isinstance(value, NamedElement):
                value.name = key
            return

        if isinstance(value, Ref):
            self._refs[key] = value

        if key == 'RE_KEYWORDS':
            if self._has_keywords:
                raise ReKeywordsChangedError(
                    'RE_KEYWORDS must be set on top of Grammar before '
                    'keywords are set.')
            self._RE_KEYWORDS = value

        if isinstance(value, NamedElement):
            if hasattr(value, 'name'):
                raise NameAssignedError(
                    'Element name is set to {0!r} and therefore cannot be '
                    'set to {1!r}. Use Repeat({0}, 1, 1) as a workaround.'
                    .format(value.name, key))
            self._check_keywords(value)
            value.name = key

        super().__setitem__(key, value)


class _OrderedClass(type):

    @classmethod
    def __prepare__(mcs, name, bases):
        return _KeepOrder()

    def __new__(mcs, name, bases, attrs, **kwargs):
        for n, ref in attrs['_refs'].items():
            if ref.element is None:
                raise MissingRefError(
                    'Forward reference {!r} is createad but the '
                    'actual reference is missing.'.format(n))
        return super().__new__(mcs, name, bases, attrs)


class Grammar(metaclass=_OrderedClass):

    __slots__ = ('_element', '_string', '_expecting', '_cached_kw_match')

    RE_LEFT_WHITESPACE = re.compile('^\s+')
    RE_KEYWORDS = _RE_KEYWORDS
    RE_WHITESPACE = re.compile('\s+')

    JS_IDENTATION = ' ' * 4
    JS_MODULE_NAME = 'jsleri'
    JS_TEMPLATE = '''
/* jshint newcap: false */

/*
 * This grammar is generated using the Grammar.export_js() method and
 * should be used with the {js_module} JavaScript module.
 *
 * Source class: {name}
 * Created at: {datetime}
 */

'use strict';

(function (
{arguments}
{ident}{ident}) {{
{language}

{ident}window.{name} = Grammar(START, '{re_keywords}');

}})(
{constructors}
);
'''.lstrip()

    PY_IDENTATION = ' ' * 4
    PY_MODULE_NAME = 'pyleri'
    PY_TEMPLATE = '''
"""
 This grammar is generated using the Grammar.export_py() method and
 should be used with the {py_module} python module.

 Source class: {name}
 Created at: {datetime}
"""
import re
{imports}

class {name}(Grammar):
{ident}
{ident}RE_KEYWORDS = re.compile('{re_keywords}')
{language}
'''.lstrip()

    C_IDENTATION = ' ' * 4
    C_TARGET = 'grammar'
    C_TEMPLATE_C = '''
/*
 * This grammar is generated using the Grammar.export_c() method and
 * should be used with the cleri module.
 *
 * Source class: {name}
 * Created at: {datetime}
 */

#include <{target}.h>
#include <stdio.h>

#define CLERI_CASE_SENSITIVE 0
#define CLERI_CASE_INSENSITIVE 1

#define CLERI_FIRST_MATCH 0
#define CLERI_MOST_GREEDY 1

cleri_grammar_t * compile_grammar(void)
{{
{language}

    cleri_grammar_t * grammar = cleri_grammar(START, "{re_keywords}");

    return grammar;
}}
'''.lstrip()

    C_TEMPLATE_H = '''
/*
 * This grammar is generated using the Grammar.export_c() method and
 * should be used with the cleri module.
 *
 * Source class: {name}
 * Created at: {datetime}
 */

#pragma once

#include <{target}.h>
#include <cleri/object.h>

cleri_grammar_t * compile_grammar(void);

enum cleri_grammar_ids {{
    CLERI_NONE,   // used for objects with no name
{enums}
    CLERI_END // can be used to get the enum length
}};

'''.lstrip()

    def __init__(self):
        '''Initialize the grammar.

        Note: usually you should only initialize a Grammar instance
        once in a project.
        '''
        self._element = self.START
        self._string = None
        self._expecting = None
        self._cached_kw_match = {}

    def export_js(
            self,
            js_module_name=JS_MODULE_NAME,
            js_template=JS_TEMPLATE,
            js_identation=JS_IDENTATION):
        '''Export the grammar to a JavaScript file which can be
        used with the js-lrparsing module.'''

        language = []
        classes = {'Grammar'}
        ident = 0

        for name in self._order:
            elem = getattr(self, name, None)
            if not isinstance(elem, Element):
                continue
            if not hasattr(elem, '_export_js'):
                continue
            language.append('{ident}var {name} = {value};'.format(
                ident=js_identation,
                name=name,
                value=elem._export_js(js_identation, ident, classes)))

        for name, ref in self._refs.items():
            language.append('{ident}Object.assign({name}, {value});'
                .format(
                    ident=js_identation,
                    name=name,
                    value=ref._element._export_js(
                        js_identation,
                        ident,
                        classes)))


        return js_template.format(
            name=self.__class__.__name__,
            ident=js_identation,
            js_module=js_module_name,
            datetime=time.strftime('%Y-%m-%d %H:%M:%S', time.localtime()),
            language='\n'.join(language),
            arguments=',\n'.join(map(lambda s:
                                     js_identation * 3 + s, classes)),
            re_keywords=self.RE_KEYWORDS.pattern.replace('\\', '\\\\'),
            constructors=',\n'.join(
                map(lambda s: js_identation + s,
                    ['.'.join(
                        [
                            'window',
                            js_module_name, n]) for n in classes])))

    def export_py(
            self,
            py_module_name=PY_MODULE_NAME,
            py_template=PY_TEMPLATE,
            py_identation=PY_IDENTATION):
        '''Export the grammar to a python file which can be
        used with the pyleri module. This can be useful when python code
        if used to auto-create a grammar and an export of the final result is
        required.'''

        language = []
        classes = {'Grammar'}
        ident = 0

        for name in self._order:
            elem = getattr(self, name, None)
            if not isinstance(elem, Element):
                continue
            if not hasattr(elem, '_export_py'):
                continue
            language.append('{ident}{name} = {value}'.format(
                ident=py_identation,
                name=name,
                value=elem._export_py(py_identation, ident, classes)))

        for name, ref in self._refs.items():
            language.append('{ident}{name} = {value}'
                .format(
                    ident=py_identation,
                    name=name,
                    value=ref._element._export_py(
                        py_identation,
                        ident,
                        classes)))

        return py_template.format(
            name=self.__class__.__name__,
            ident=py_identation,
            py_module=py_module_name,
            datetime=time.strftime('%Y-%m-%d %H:%M:%S', time.localtime()),
            language='\n'.join(language),
            re_keywords=self.RE_KEYWORDS.pattern.replace('\\', '\\\\'),
            imports='\n'.join(
                map(lambda s: s,
                    [' '.join(['from', py_module_name, 'import', n])
                    for n in classes if n != 'Rule'])))

    def export_c(self, target=C_TARGET, c_identation=C_IDENTATION):
        '''Export the grammar to a c (source and header) file which can be
        used with the cleri module.'''
        language = []
        ident = 0
        enums = set()
        for name in self._order:
            elem = getattr(self, name, None)
            if not isinstance(elem, Element):
                continue
            if not hasattr(elem, '_export_c'):
                continue
            if isinstance(elem, Ref):
                language.append(
                    '{ident}cleri_object_t * {name};'.format(
                        ident=c_identation,
                        name=name))
            else:
                language.append(
                    '{ident}cleri_object_t * {name} = {value};'.format(
                        ident=c_identation,
                        name=name,
                        value=elem._export_c(c_identation, ident, enums)))

        for name, ref in self._refs.items():
            language.append('{ident}{name} = {value};'
                .format(
                    ident=c_identation,
                    name=name,
                    value=ref._element._export_c(
                        c_identation,
                        ident,
                        enums)))

        pattern = self.RE_KEYWORDS.pattern.replace('\\', '\\\\')
        if not pattern.startswith('^'):
            pattern = '^' + pattern

        enums = ',\n'.join([
            '{}{}'.format(c_identation, gid)
            for gid in sorted(enums)]) + ','

        return (self.__class__.C_TEMPLATE_C.format(
                    name=self.__class__.__name__,
                    target=target,
                    ident=c_identation,
                    datetime=time.strftime('%Y-%m-%d %H:%M:%S', time.localtime()),
                    language='\n'.join(language),
                    re_keywords=pattern),
                self.__class__.C_TEMPLATE_H.format(
                    name=self.__class__.__name__,
                    target=target,
                    datetime=time.strftime('%Y-%m-%d %H:%M:%S', time.localtime()),
                    language='\n'.join(language),
                    enums=enums))

    def parse(self, string):
        '''Parse some string to the Grammar.

        Returns a nodeResult with the following attributes:
         - is_valid: True when the string is successfully parsed
                     by the Grammar.
         - pos: position in the string where parsing ended.
                (this is the end of the string when is_valid is True)
         - expecting: a list containing possible elements at position
                      'pos' in the string.
         - tree: the parse_tree containing a structured
                 result for the given string.
        '''
        self._string = string
        self._expecting = Expecting()
        self._cached_kw_match.clear()
        self._len_string = len(string)
        self._pos = None
        tree = Node(self._element, string, 0, self._len_string)
        node_res = NodeResult(*self._walk(
            self._element,
            0,
            tree.children,
            self._element,
            True))

        # get rest if anything
        rest = self._string[node_res.pos:].lstrip()

        # set is_valid to False if we have 'rest' left.
        if node_res.is_valid and rest:
            node_res.is_valid = False

        # add end_of_statement to expecting if this is possible
        if not self._expecting.required and rest:
            self._expecting.set_mode_required(node_res.pos, True)
            self._expecting.update(end_of_statement, node_res.pos)

        node_res.expecting = self._expecting.get_expecting()

        # add expecting and correct pos to node_res if node_res is not valid
        if not node_res.is_valid:
            node_res.pos = self._expecting.pos

        node_res.tree = tree

        return node_res

    def _append_tree(self, tree, node, pos):
        if pos > self._expecting.pos:
            self._expecting.empty()
        node.end = pos
        tree.append(node)

    def _walk(self, element, pos, tree, rule, is_required):
        if self._pos != pos:
            self._s = self._string[pos:].lstrip()
            self._pos = self._len_string - len(self._s)
        node = Node(element, self._string, self._pos)
        self._expecting.set_mode_required(node.start, is_required)

        return element._get_node_result(self, tree, rule, self._s, node)



'''NodeResult class.

:copyright: 2015, Jeroen van der Heijden (Transceptor Technology)
'''


class NodeResult:

    __slots__ = ('is_valid', 'pos', 'expecting', 'tree')

    def __init__(self, is_valid, pos):
        self.is_valid = is_valid
        self.pos = pos
        self.expecting = None
        self.tree = None



'''pyleri.Keyword Class.

Try matching a given keyword string. The keyword should match
Grammer.RE_KEYWORDS otherwise the keyword will not be found. It's possible
however to overwrite the default RE_KEYWORDS in you own Grammar class.

Example:

class MyGrammar(Grammar):

    # needed because the default ^\w+ will not match keyword: my-keyword
    RE_KEYWORDS = re.compile('^[\w-]+')

    my_keyword = Keyword('my-keyword')
    ...


:copyright: 2015, Jeroen van der Heijden (Transceptor Technology)
'''
from .elements import NamedElement, c_export


class Keyword(NamedElement):

    __slots__ = ('_keyword', '_ign_case')

    def __init__(self, keyword, ign_case=False):

        if not isinstance(keyword, str):
            raise TypeError(
                'Keyword(): first positional argument must be a string value')

        self._keyword = keyword
        self._ign_case = bool(ign_case)

    def __repr__(self):
        return self._keyword

    def _get_node_result(self, root, tree, rule, s, node):
        try:
            re_match = root._cached_kw_match[node.start]
        except KeyError:
            re_match = \
                root._cached_kw_match[node.start] = \
                root.RE_KEYWORDS.match(s)
        is_valid = \
            re_match and \
            (re_match.group(0) == self._keyword or
             (self._ign_case and
              re_match.group(0).lower() == self._keyword.lower()))
        if is_valid:
            root._append_tree(tree, node, node.start + len(self._keyword))
        else:
            root._expecting.update(self, node.start)

        return is_valid, node.end or node.start

    def _run_export_js(self, js_identation, ident, classes):
        return 'Keyword(\'{}\'{})'.format(
            self._keyword,
            ', true' if self._ign_case else '')

    def _run_export_py(self, py_identation, ident, classes):
        return 'Keyword(\'{}\'{})'.format(
            self._keyword,
            ', True' if self._ign_case else '')

    @c_export
    def _run_export_c(self, c_identation, ident, enums, gid):
        return 'cleri_keyword({}, "{}", {})'.format(
            gid,
            self._keyword,
            'CLERI_CASE_INSENSITIVE'
            if self._ign_case else 'CLERI_CASE_INSENSITIVE')



'''pyleri.Choice Class.

Choose one of the given elements. When most_greedy is True we will choose
the 'longest' element when multiple elements are valid. If most_greedy is
False we will return the first match.

:copyright: 2015, Jeroen van der Heijden (Transceptor Technology)
'''

from .elements import NamedElement, c_export


class Choice(NamedElement):

    __slots__ = ('_elements', '_get_node_result')

    def __init__(self, *elements, most_greedy=True):
        self._elements = self._validate_elements(elements)
        self._get_node_result = \
            self._most_greedy_result if most_greedy else \
            self._stop_at_first_match

    def _most_greedy_result(self, root, tree, rule, s, node):
        mg_is_valid, mg_pos = False, node.start

        for elem in self._elements:
            children = []
            is_valid, pos = root._walk(elem, node.start, children, rule, True)

            if is_valid and pos > mg_pos:
                node.children = children
                mg_is_valid, mg_pos = is_valid, pos

        if mg_is_valid:
            root._append_tree(tree, node, mg_pos)

        return mg_is_valid, mg_pos

    def _stop_at_first_match(self, root, tree, rule, s, node):
        for elem in self._elements:
            children = []
            is_valid, pos = root._walk(elem, node.start, children, rule, True)

            if is_valid:
                node.children = children
                root._append_tree(tree, node, pos)
                break

        return is_valid, pos

    def _run_export_js(self, js_identation, ident, classes):
        return self._export_js_elements(js_identation, ident, classes)

    def _run_export_py(self, py_identation, ident, classes):
        new_ident = ident + 1
        value = ',\n'.join(['{ident}{elem}'.format(
            ident=py_identation * new_ident,
            elem=elem._export_py(
                py_identation,
                new_ident,
                classes)) for elem in self._elements])
        return 'Choice(\n{val},\n{ident}most_greedy={mg})'.format(
            mg='{mg}'.format(
                ident=py_identation * (ident + 1),
                mg=('False', 'True')[
                    self._get_node_result == self._most_greedy_result]),
            val=value,
            ident=py_identation * new_ident)

    @c_export
    def _run_export_c(self, c_identation, ident, enums, gid):
        new_ident = ident + 1
        value = ',\n'.join(['{ident}{elem}'.format(
            ident=c_identation * new_ident,
            elem=elem._export_c(
                c_identation,
                new_ident,
                enums)) for elem in self._elements])
        return 'cleri_choice(\n{gid},\n{mg},\n{num},\n{val}\n{ident})'.format(
            gid='{ident}{gid}'.format(
                ident=c_identation * (ident + 1),
                gid=gid),
            mg='{ident}{mg}'.format(
                ident=c_identation * (ident + 1),
                mg=('CLERI_FIRST_MATCH', 'CLERI_MOST_GREEDY')[
                    self._get_node_result == self._most_greedy_result]),
            num='{ident}{num}'.format(
                ident=c_identation * (ident + 1),
                num=len(self._elements)),
            val=value,
            ident=c_identation * ident)



'''Tokens class.

:copyright: 2015, Jeroen van der Heijden (Transceptor Technology)
'''
from .elements import NamedElement, c_export


class Tokens(NamedElement):

    __slots__ = ('_tokens',)

    def __init__(self, tokens):
        self._tokens = tokens.split()
        self._tokens.sort(key=len, reverse=True)

    def __repr__(self):
        return ', '.join(self._tokens)

    def _get_node_result(self, root, tree, rule, s, node):
        for token in self._tokens:
            if s.startswith(token):
                root._append_tree(tree, node, node.start + len(token))
                return True, node.end

        root._expecting.update(self, node.start)

        return False, node.start

    def _run_export_js(self, js_identation, ident, classes):
        return 'Tokens(\'{}\')'

    def _run_export_py(self, py_identation, ident, classes):
        return 'Tokens(\'{}\')'

    @c_export
    def _run_export_c(self, c_identation, ident, enums, gid):
        return 'cleri_tokens({}, "{}")'.format(
            gid,
            ' '.join(self._tokens))



'''This class and THIS instance.

:copyright: 2015, Jeroen van der Heijden (Transceptor Technology)
'''
from .elements import Element


class This(Element):

    def _get_node_result(self, root, tree, rule, _s, node):
        if node.start not in rule._tested:
            rule._tested[node.start] = root._walk(
                rule._element,
                node.start,
                node.children,
                rule,
                True)
            rule._tree[node.start] = node.children

        if rule._tested[node.start][0]:
            node.children = rule._tree[node.start]
            root._append_tree(tree, node, rule._tested[node.start][1])

        return rule._tested[node.start]

    def _export_js(self, js_identation, ident, classes):
        classes.add('THIS')
        return 'THIS'

    def _export_py(self, py_identation, ident, classes):
        classes.add('THIS')
        return 'THIS'

    def _export_c(self, c_identation, ident, enums):
        return 'CLERI_THIS'

THIS = This()



import unittest
import gc
import sys
sys.path.append('..')

from pyleri import (
    Keyword,
    Sequence,
    Optional,
    Prio,
    Choice,
    THIS,
    Ref,
    List,
    Grammar,
    KeywordError,
    ReKeywordsChangedError,
    NameAssignedError)
from examples import JsonGrammar


class _TestGrammar1(Grammar):
    k_test = Keyword('test')
    k_ignore_case = Keyword('ignore_case', ign_case=True)
    START = Sequence(k_test, k_ignore_case)


class _TestGrammar2(Grammar):
    k_ni = Keyword('ni')
    s_seq = Sequence('(', THIS, ')')
    START = Prio(
        k_ni,
        s_seq,
        Sequence(THIS, Keyword('and'), THIS),
        Sequence(THIS, Keyword('or'), THIS))


class _TestGrammar3(Grammar):
    s_tic_tac = Sequence(Keyword('tic'), Keyword('tac'))
    s_tic_tac_toe = Sequence(Keyword('tic'), Keyword('tac'), Keyword('toe'))

    START = Sequence(
        Choice(s_tic_tac, s_tic_tac_toe),
        Choice(s_tic_tac, s_tic_tac_toe, most_greedy=False))


class _TestGrammar4(Grammar):
    START = Ref()
    ni_item = Choice(Keyword('ni'), START)
    START = Sequence('[', List(ni_item), ']')


class TestPyleri(unittest.TestCase):

    def setUp(self):
        gc.collect()

    def test_parse_keyword(self):
        tg = _TestGrammar1()
        self.assertTrue(tg.parse('test ignore_case').is_valid)
        self.assertTrue(tg.parse('test Ignore_Case').is_valid)
        # test is not case insensitive
        self.assertFalse(tg.parse('Test ignore_case').is_valid)
        # - is not _
        self.assertFalse(tg.parse('test ignore-case').is_valid)

    def test_noderesult_from_sequence(self):
        tg = _TestGrammar1()
        k_test, k_ignore_case = tg.parse(
                'test ignore_case').tree.children[0].children
        # element should be the original element
        self.assertEqual(k_test.element, _TestGrammar1.k_test)

        # Keyword string should be equal to str(Keyword)
        self.assertEqual(k_test.string, str(_TestGrammar1.k_test))

        # Test second node result
        self.assertEqual(k_ignore_case.element, _TestGrammar1.k_ignore_case)

    def test_compile_error(self):
        with self.assertRaises(ReKeywordsChangedError):
            class _G1(Grammar):
                k_test = Keyword('test')
                RE_KEYWORDS = '[a-z]+'

        with self.assertRaises(KeywordError):
            class _G2(Grammar):
                k_test = Keyword('test-nomatch')

        with self.assertRaises(KeywordError):
            # test 'deep' keyword checks.
            class _G3(Grammar):
                START = Optional(Keyword('test-nomatch'))

        with self.assertRaises(NameAssignedError):
            class _G4(Grammar):
                k_test = Keyword('test')
                k_test2 = k_test

    def test_prio(self):
        tg = _TestGrammar2()
        self.assertTrue(tg.parse('ni').is_valid)
        self.assertTrue(tg.parse('(ni)').is_valid)
        self.assertTrue(tg.parse('ni and ni').is_valid)
        self.assertTrue(tg.parse('((ni) and ni and (ni or ni))').is_valid)
        # missing closing )
        self.assertFalse(tg.parse('(ni or ni and ni').is_valid)
        # missing and/or between ni and (ni)
        self.assertFalse(tg.parse('ni (ni)').is_valid)

    def test_choice(self):
        tg = _TestGrammar3()
        # both should stop at first match
        self.assertTrue(tg.parse('tic tac tic tac').is_valid)
        # tic tac toe should be matched because most_greede is left default
        # (True) for the first choice
        self.assertTrue(tg.parse('tic tac toe tic tac').is_valid)
        # the second tic tac toe should NOT be matched because most_greede is
        # set to False for the second choice
        self.assertFalse(tg.parse('tic tac toe tic tac toe').is_valid)

    def test_ref(self):
        tg = _TestGrammar4()
        # should be true
        self.assertTrue(tg.parse('[ni, ni, [ni, [], [ni, ni]]]').is_valid)

    def test_json(self):
        tg = JsonGrammar()
        # should be true
        self.assertTrue(tg.parse('''
            {
                "hoi \\"Iris\\"": [
                    5,
                    -0.5,
                    {
                        "a": true,
                        "b": false
                    }
                ]
            }''').is_valid)

        # keys must be strings
        self.assertFalse(tg.parse('{5: true}').is_valid)

        # floats should start with 0-9.
        self.assertFalse(tg.parse('{"a": .5}').is_valid)

        # wrong escaping
        self.assertFalse(tg.parse('{"a\\": null}').is_valid)

        # empty strings are not allowed
        self.assertFalse(tg.parse('').is_valid)

    def tearDown(self):
        self.assertEqual(gc.collect(), 0, msg=self.id())


if __name__ == '__main__':
    unittest.main()

    # grammar = JsonGrammar()
    # print(grammar.export_c()[0])


'''Rule class.

:copyright: 2015, Jeroen van der Heijden (Transceptor Technology)
'''
from .elements import NamedElement


class Rule(NamedElement):

    __slots__ = ('_element', '_tested', '_tree')

    def __init__(self, element):
        self._element = element

    def _get_node_result(self, root, tree, rule, _s, node):
        self._tested = {}
        self._tree = {}
        is_valid, pos = root._walk(
            self._element,
            node.start,
            node.children,
            self,
            True)
        if is_valid:
            root._append_tree(tree, node, pos)
        return is_valid, pos

    def _run_export_js(self, js_identation, ident, classes):
        return self._element._export_js(js_identation, ident, classes)

    def _run_export_py(self, py_identation, ident, classes):
        return self._element._export_py(py_identation, ident, classes)

    def _run_export_c(self, c_identation, ident, enums):
        name = getattr(self, 'name', None)
        if name is not None:
            self._element._name = name
        return self._element._export_c(c_identation, ident, enums)



from distutils.core import setup

VERSION = '1.1.3'

setup(
    name='pyleri',
    packages=['pyleri'],
    version=VERSION,
    description='Python Left-Right Parser',
    author='Jeroen van der Heijden',
    author_email='jeroen@transceptor.technology',
    url='https://github.com/transceptor-technology/pyleri',
    download_url=
        'https://github.com/transceptor-technology/'
        'pyleri/tarball/{}'.format(VERSION),
    keywords=['parser', 'grammar', 'autocompletion'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.0',
        'Programming Language :: Python :: 3.1',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Text Processing :: Linguistic'
    ],
)
